# Binary surface plot
import numpy as np
import matplotlib.pyplot as plt

def binaire(i,j,beta,*args): # creates a model with beta coefficients : [i,j,ij,i2,j2]

    # Use this model to cover all the space
    I = []
    J = []
    
    for ii in np.arange(min(i),max(i),(max(i)-min(i))/100):
        for jj in np.arange(min(j),max(j),(max(j)-min(j))/100):
            I.append(ii)
            J.append(jj)
         
    I = np.array(I)[np.newaxis].T
    J = np.array(J)[np.newaxis].T
    
    IJ = I*J
    I2 = I*I
    J2 = J*J
    onez = np.ones((len(I),1))
    X = np.concatenate((onez,I,J,IJ,I2,J2),axis=1)

    yhat = X@beta
 
    plt.figure()
    plt.scatter(I,J,marker='s',c=yhat,cmap=plt.cm.plasma)
    try : plt.xlabel(args[0],fontsize=16)
    except : plt.xlabel('$x_i$',fontsize=16)
    try : plt.ylabel(args[1],fontsize=16)
    except : plt.ylabel('$x_j$',fontsize=16)  
    plt.grid('on',c='black')    
    plt.xlim(min(i),max(i))
    plt.ylim(min(j),max(j))
    plt.colorbar()
    plt.tight_layout()




# PCA: principal component analysis

import numpy as np

def PCA(X, nbPC):
    ''' 
    
    PCA: principal component analysis
    [T,P,SSX] = PCA(X, nbPC)
    
    INPUT
    npPC [1 x 1] <numpy.ndarray>
        number of components in the PCA decomposition
    X [n x k] <numpy.ndarray>
        spectra
        n samples
        k variables
    
    OUTPUT
    T [n x nbPC] 
        scores
    P [nbPC x k] 
        loadings
    SSX [1 x nbPC] 
        Sum of squared variance explained by each component
        
    '''
    
    s0, s1 = X.shape
    T = np.zeros((s0, nbPC))
    P = np.zeros((s1, nbPC))
    SSX = np.zeros((nbPC, 1))
    X0 = X # Save a copy
    
    for i in range(0,nbPC):
        error = 1
        t_temp = np.ones((s0, 1))
        t = np.ones((s0, 1))
            
        while error > 1E-10:
            p = (X.T@t)/(t.T@t)
            p = p/np.linalg.norm(p,2)

            # Fix rotational ambiguity
            onez = np.ones([len(p),1]) # Use as a reference
            if p.T@onez < 0:
                p = -p

            t = (X@p)/(p.T@p)
            # Check t convergence --------------------
            error = sum(np.power((t-t_temp),2),0) # Squared error
            t_temp = t
            # ----------------------------------------     
        P[:,i] = np.squeeze(p)
        T[:,i] = np.squeeze(t)
        X = X - t@p.T
        
        # Sum of squares -----------
        Xhat = t@p.T
        ssX0 = np.sum(np.sum(X0 * X0)) # Each element squared
        ssX = np.sum(np.sum(Xhat * Xhat))
        ssX = ssX / ssX0
        SSX[i] = ssX
        # --------------------------    
        
    return T, P, SSX



# ANOVA 
# Ryan Gosselin
# Version 2023

import statsmodels.api as sm
from statsmodels.formula.api import ols
import numpy as np
import pandas as pd



def get_index_positions(list_of_elems, element):
    ''' Returns the indexes of all occurrences of give element in
    the list- listOfElements '''
    index_pos_list = []
    index_pos = 0
    while True:
        try:
            # Search for item in list from indexPos to the end of list
            index_pos = list_of_elems.index(element, index_pos)
            # Add the index position in list
            index_pos_list.append(index_pos)
            index_pos += 1
        #except ValueError as e:
        except:
            break
    return index_pos_list



def include_C(text):
    # OLS expects categorical variables in ANOVA
    # to do so, add C() aroun each variable

    # Find + 
    plus = get_index_positions(text, '+')
    plus = np.array(plus)[np.newaxis]
    # Find ~
    wave = get_index_positions(text, '~')
    wave = np.array(wave)[np.newaxis]
    # Find :
    dots = get_index_positions(text, ':')
    dots = np.array(dots)[np.newaxis]
    # How long is the text string
    length = len(text)
    length = np.array(length)[np.newaxis][np.newaxis]
    # Combine all these items
    parts = np.concatenate((plus, wave, dots, length),axis=1)
    parts = np.sort(parts)

    # Add C() around each variable
    formula = ''
    formula = formula + text[0:int(parts[0,0])+1]
    #print(formula)
    
    for i in range(parts.shape[1]-1):
        word = text[int(parts[0,i])+1:int(parts[0,i+1])]
        word = 'C('+word+')'
        #print(word)
        formula = formula + word
        #print(formula) 
        try:
            sign = text[int(parts[0,i+1])]
            formula = formula + sign
        except:
            break
    return formula


def exclude_C(text):
    # OLS expects categorical variables in ANOVA
    # to do so, add C() aroun each variable
    # After OLS, remove it to show in table

   # Find + 
    plus = get_index_positions(text, '+')
    plus = np.array(plus)[np.newaxis]
    # Find ~
    wave = get_index_positions(text, '~')
    wave = np.array(wave)[np.newaxis]
    # How long is the text string
    length = len(text)
    length = np.array(length)[np.newaxis][np.newaxis]
    # Combine all these items
    parts = np.concatenate((plus, wave, length),axis=1)
    parts = np.sort(parts)

    WORDS = []
    for i in range(parts.shape[1]-1):
        word = text[int(parts[0,i])+1:int(parts[0,i+1])]
        WORDS.append(word)
    return WORDS

def anova(data,text,general):
    
    '''
    ANOVA
    
    data: données dans un format pandas
    text: modèle ANOVA recherché
    general: 1 si un plan d'experience ANOVA à plusieurs niveaux (3 et +)
    '''
    
    # OLS expects categorical variables in ANOVA
    # to do so, add C() aroun each variable
    if general == 1:
        formula = include_C(text)
    else:
        formula = text
      
    model = ols(formula, data).fit()
    table = sm.stats.anova_lm(model, typ=2)
    
    MS = table['sum_sq']/table['df'] 
    table.insert(2, 'MS', MS)
    
    table.columns = ['SS', 'df', 'MS', 'F','p']
    
    WORDS = exclude_C(text)
    WORDS.append('Erreur')
    WORDS.append('Total')
    
    
    # Totals
    y_bar = np.mean(data.y)
    SST = np.sum((data.y - y_bar)**2)
    dfT = len(data.y)-1
    
    
    # Check for confounding
    # Summed within the table
    #SSsum = table['SS'].sum() 
    dfsum = table['df'].sum()
    
    
    # Include row for totals
    d_row = [SST,dfT,'-','-','-']
    table.loc[len(table)] = d_row
    
    table.index = WORDS
    
    # Get rid of NaN
    table.at['Erreur', 'F'] = ('-')
    table.at['Erreur', 'p'] = ('-')
    
    freedom = dfsum > dfT
    
    if freedom == 1:
        print('\n..........................................')
        print('ATTENTION!')
        print('Il manque %.0f degré de liberté.' %freedom)    
        print('Vérifier le patron de confusion.')
        print('..........................................')
        print('\n')
    if freedom > 1:
        print('\n..........................................')
        print('ATTENTION!')
        print('Il manque %.0f degrés de liberté.' %freedom)   
        print('Vérifier le patron de confusion.')
        print('..........................................')
        print('\n')
     
    return table


# normplot: normal distribution plot
# Ryan Gosselin
# Version 2023

from scipy import stats
import matplotlib.pyplot as plt
import numpy as np
    
def normplot(x, *args):

    ''' 
    
    normplot: normal distribution plot
    [] = normplot(X, tag)
    
    INPUT
    X [n x 1] <numpy.ndarray>
        n samples
    tag [n x 1] <list>
        (optional input)
        tag for each datapoint (ie name)
    
    OUTPUT
    plot
        
    '''

    #Label points
    if len(args) != 0:
        point_labels = list(args[0])
    else:
        point_labels = list(range(0, max(x.shape))) 

    # Sort x but keep track of the order of the observations
    order = np.arange(0, len(x), 1)
    order = (order[np.newaxis]).T
    xorder = np.concatenate((x,order), axis=1)
    xorder = xorder[xorder[:,0].argsort()]
    x = xorder[:,0]
    order = xorder[:,1]
    o = []
    for i in range(len(order)):
        o.append(int(order[i]))
    
    point_labels = [point_labels[i] for i in o]


    x = np.squeeze(x)
    
    # Calculate quantiles and least-square-fit curve
    (quantiles, values), (slope, intercept, r) = stats.probplot(x, dist='norm')
    
    plt.figure()  
    plt.plot(values, quantiles,'.w')
        
     
    
    for i, label in enumerate(point_labels): 
        plt.text (values[i], quantiles[i], label, fontsize=15).set_color('blue') 
    
    
     
    plt.plot(quantiles * slope + intercept, quantiles, 'r',linewidth=0.5)
    
    
    #define ticks
    ticks_perc=[1, 5, 10, 20, 50, 80, 90, 95, 99]
    
    #transfrom them from precentile to cumulative density
    ticks_quan=[stats.norm.ppf(i/100.) for i in ticks_perc]
    
    #assign new ticks
    plt.yticks(ticks_quan,ticks_perc)

    plt.xlabel('variable',fontsize=20)
    plt.ylabel('probabilite',fontsize=20)

    #show plot
    plt.tight_layout()
    plt.show()



from sherbystats.anova import *
from sherbystats.doe import *
from sherbystats.mlr import *
from sherbystats.lda import *
from sherbystats.normplot import *
from sherbystats.pca import *
from sherbystats.pca_trou import *
from sherbystats.xlsread import *

from sherbystats.ternaire import *
from sherbystats.binaire import *
from sherbystats.binaire3D import *













# PCA trous

import numpy as np

def PCA_missing(X, nbPC):

    s0, s1 = X.shape
    T = np.zeros((s0, nbPC))
    P = np.zeros((s1, nbPC))
        
    for k in range(0,nbPC):
        error = 1
        t_temp = np.ones((s0, 1))
        t = np.ones((s0, 1))
        
        while error > 1E-6:
            
            # The next few lines do: 
            # p = (X.T@t)/(t.T@t)
            # p = p/np.linalg.norm(p,2)
            num = np.zeros((1,X.shape[1]))
            denom = np.zeros((1,X.shape[1]))
            for i in range(X.shape[0]):
                Xt = X[i,:]*t[i]
                Xt[np.isnan(Xt)] = 0 # Replace NaNs by 0
                num = num + Xt
                tt = t[i]**2
                denom = denom + tt 
            p = (num / denom).T
            p = p/np.linalg.norm(p,2)
            # np.linalg.norm(p,2) = np.max(np.sqrt(p.T @ p))
    
            # The next few lines do: 
            # t = (X@p)/(p.T@p)
            num = np.zeros((1,X.shape[0]))
            denom = np.zeros((1,X.shape[0]))
            
            for j in range(X.shape[1]):
                Xp = X[:,j]*p[j]
                Xp[np.isnan(Xp)] = 0 # Replace NaNs by 0
                num = num + Xp
                
                pp = p[j]**2
                denom = denom + pp
            t = (num / denom).T
    
            # Check t convergence --------------------
            error = sum(np.power((t-t_temp),2),0) # Squared error
            t_temp = t
            # ----------------------------------------     
        P[:,k] = np.squeeze(p)
        T[:,k] = np.squeeze(t)
        X = X - t@p.T
    
    return T,P        